import openmdao.api


def extract_pareto(
    prob: openmdao.api.Problem,
    ard_input_dict: dict,
):

    # grab the driver-scaled values off of the drivers
    desvar_nd = prob.driver.desvar_nd.copy()
    obj_nd = prob.driver.obj_nd.copy()

    #
